//********************************************************************************************    
/* FA-Leserservice, 01.11.2023
	  Die Software lädt einen DRA-818V (Empfänger) und einen DRA818U (Sender) mit den
    Betriebsparametern, u.a. dem Betrag der Arbeitsfrequenz. Die TRX-Module bilden einen
    Crossband-Repeater von 2m auf 70cm.
    Der Squelch-Ausgang des DRA-818V wird überwacht. Sobald er auf LOW-Pegel geht, wird
    der PTT-Anschluss des DRA-818U ebenfalls auf LOW gesetzt, um den Sender einzuschalten.
    Dieser sendet dann das vom 2-m-Empfänger kommende Signal auf einer Frequenz im 70-cm-Band.
    Wenn der Squelch-Ausgang des DAR-818V wieder auf HIGH-Pegel geht, wird der UHF-Sender mit 
    Verzögerung wieder abgeschaltet. Die Verzögerungszeit ist per Software einstellbar.
    Überwachungsfunktionen: - Versorgungsspannung 12V
                            - Kühlkörpertemperatur der Endstufe
                            - Ausgangspeger der Endstufe
    Bei Abweichungen vom Sollwert werden entsprechende Schutzfunktionen wirksam, die 
    inbesondere die Endstufe vor Zerstörung schützen sollen. Dies soll einen möglichst langen
    autarken Betrieb sicherstellen. Das Auftreten der Störungen wird durch entsprechende Kürzel
    im Morsecode auf der Sendefreuquenz signalisiert.
    Weitere Details siehe Programmablaufplan in der Bauanleitung des Crossband-Repeater-Bausatzes.
	
Versionsnummer: 1.0
  Grundversion

*/

//******************************************************************************************
/*Anschlussbelegung des Arduino Nano:  
  TXD   Erweiterungsschnittstelle UART-Ausgang  
  RXD   Erweiterungsschnittstelle UART-Eingang  
	D2		Eingang Squelch DRA-818V
	D3		Reserve, Erweiterungsschnittstelle
	D4		Ausgang für PTT DRA-818U
	D5		Power-Down-Ausgang für DRA-818U
	D6		Power-Down-Ausgang für DRA-818V
	D7		Steuerausgang RXU
	D8		Steuerausgang RXV
	D9		Steuereingang TXU
	D10		Steuereingang TXV
	D11  	Ausgang für Tonsignal
	D12 	Schaltausgang Bypassrelais PA-Modul
	D13		Schaltausgang Gatespannung PA-Modul
	A0		Messeingang Akkuspannung 12V
  A1    Eingang Steuerspannung zur Squelch-Einstellung
  A2    LED-Schaltausgang Betrieb/Störung
  A3    LED-Schaltausgang Sendekontrolle (Reserveschaltstufe PA-Modul)
  A4    Erweiterungsschnittstelle I2C/SDA
  A5    Erweiterungsschnittstelle I2C/SCL
  A7    Eingang Temperaturmessfühler PA-Modul
  A6    Messeingang HF-Spannung am Ausgang der PA
  VIN		Betriebsspannungseingang 8V
	GND		Masse
*/

//***********************************************
// Bibliothek für die serielle Softwareschnittstelle einbinden

#include <SoftwareSerial.h>

//Definition von Variablen und Konstanten

//Empfangsfrequenz des VHF-Moduls, 
//Wert zwischen 144.0000 und 146.0000:
const float FRX_V = 145.1000;   

//Sendefrequenz des UHF-Moduls,
//Wert zwischen 430.0000 und 440.0000:
const float FTX_U = 431.5000;     

//**********************************************************************************************************************************************************

//Festlegung weiterer Betriebsparameter für den DRA-818V
const int GBW_V = 1;               //Bandbreite des VHF-Moduls; 1 -> 25 kHz
const String TX_CTCSS_V = "0000";  //CTCSS-Frequenz TX; (wird hier nur der Vollständikeit halber mitgesendet); Wert zwischen 0000 (CTCSS "aus") und 0038
const String RX_CTCSS_V = "0000";  //CTCSS-Frequenz RX; (wird hier nur der Vollständikeit halber mitgesendet); Wert zwischen 0000 (CTCSS "aus") und 0038
const int VOL_V = 4;               //NF-Lautstärkeeinstellung; Wert zwischen 1 und 8

//Festlegung weiterer Betriebsparameter für den DRA-818U
const int GBW_U = 1;               //Bandbreite des VHF-Moduls; 1 -> 25 kHz
const String TX_CTCSS_U = "0000";  //CTCSS-Frequenz TX; (wird hier nur der Vollständikeit halber mitgesendet); Wert zwischen 0000 (CTCSS "aus") und 0038
const String RX_CTCSS_U = "0000";  //CTCSS-Frequenz RX; (wird hier nur der Vollständikeit halber mitgesendet); Wert zwischen 0000 (CTCSS "aus") und 0038
const int SQU_U = 8;              //Pegel der Rauschsperre, Wert zwischen 0 (offen) und 8
const int VOL_U = 1;              //NF-Lautstärke, hier auf Minimum


//Weitere Konstanten

const int TX_DELAY = 2000;  //Abschaltverzögerung des Senders beim Umschalten auf Empfang, Wert in Millisekunden
const int PA_OFF_DELAY = 10000; //Intervall nach der Fehlerabschaltung der PA bis zum erneuten Zuschaltversuch, Wert in Millisekunden
const int DIT_SPEED = 80; //Ausgabegeschwindigkeit der Morsezeichen
const int TON_HZ = 600; //Morsezeichentonhöhe in Hertz
const unsigned long MORSEINTERVALL = 20000; //Ausgabeintervall der Morsezeichen zur Fehlermeldung, Wert in Millisekunden
const int UB12_MIN = 10; //Untere Betriebsspannungsschwelle in Volt ohne Nachkommastelle
const int UB12HYST = 1; //Hysteresewert zum Zurücksetzen der Unterspannungsmeldung
const int TMP_BEZ = 792; //Bezugswert für 0 Grad Celsius zur Temperaturumrechnung 
const int TMP_S1 = 45; //Schwellwert für erhöhte Kühlkörpertemperatur in Grad Celsius ohne Nachkommastelle
const int TMP_S2 = 55; //Schwellwert für Kühlkörper-Übertemperatur in Grad Celsius ohne Nachkommastelle
const int TMPHYST = 5; //Hysteresewert in Grad Celsius zum Zurücksetzen der Übertemperaturmeldung
const int PA_UMIN = 400; //Minimalwert der Ausgangsspannung der Endstufe, "400" entspricht etwa 4 W
String meldung; //allgemeine Stringvariable


//Variablen und Anfangswerte 
int squ_v = 0;    //Speichervariable für den Pegel der Rauschsperre mit Initalwert, der letzlich wirksame Wert wird über Spannungswert an A1 definiert
int kktmp_c = 0; //Kühlkörpertemperatur in Grad Celsius
int uoutpa = 0; //Messwert HF-Ausgangsspannung am Ausgang der PA
byte statusspeicher = 0b00000000; //Variable zur Speicherung von Statusmeldungen, wird auf 1 gesetzt, wenn betreffender Status vorliegt, anderenfalls auf 0 zurückgesetzt
/*  Bit 0: Unterspannung
    Bit 1: erhöhte Kühlkörpertemperatur
    Bit 2: Übertemperatur
    Bit 3: zu geringe Ausgangsleistung, PA-Störung oder Antennenfehler
    Bit 4: (nicht belegt)
    Bit 5: (nicht belegt)
    Bit 6: (nicht belegt)
    Bit 7: Speicherbit für PTT-Status
*/
unsigned long altezeit1; //Hilfsvariable zur Zeitverzögerung beim Umschalten von Senden zurück auf Emnpfang
unsigned long altezeit2; //Hilfsvariable zur Zeitsteuerung der Morsezeichenausgabe
unsigned long altezeit3; //Hilfsvariable zur Zeitsteuerung der PA-Abschaltung bei PA-/Antennenstörung

//Zuordnung der Arduino-Pins
#define SQV 2
#define PTTU 4
#define PDU 5
#define PDV 6
#define RXU 7
#define RXV 8
#define TXU 9
#define TXV 10
#define NFOUT 11
#define BYP 12
#define UGATE 13
#define U12 A0
#define USQ A1
#define LED_rtgn A2
#define SRR A3
#define USEND A6
#define UTMP A7

// Pins zu den seriellen Schnittstellen für den Datenaustausch mit den VHF/UHF-Modulen zuordnen
SoftwareSerial Ser818V(TXV, RXV);
SoftwareSerial Ser818U(TXU, RXU);

//Debug-Monitor ein- bzw. ausschalten:
//#define debug

//Funktionsdeklaration:
void ub_messung (void);
void pa_einschalten (void);
void pa_abschalten (void);
void pa_sperre (void);
void tmp_messung (void);
void punktausgabe(void);
void strichausgabe(void);
void p_s_abstand(void);
void zeichenpause(void);
void wortpause(void);
void mz_ausgabe(char mzeichen);
void meldungsausgabe (String meldung);

//*******************************************************************
//*******************Setup-Anweisungen*******************************
//*******************************************************************
void setup() 
{
//Ein- und Ausgänge definieren:
pinMode (SQV, INPUT);
pinMode (PTTU, OUTPUT);
pinMode (PDU, OUTPUT);
pinMode (PDV, OUTPUT);
pinMode (RXU, OUTPUT);
pinMode (RXV, OUTPUT);
pinMode (TXU, INPUT);
pinMode (TXV, INPUT);
pinMode (NFOUT, OUTPUT);
pinMode (BYP, OUTPUT);
pinMode (UGATE, OUTPUT);
pinMode (LED_rtgn, OUTPUT);
pinMode (SRR, OUTPUT);

//Referenzspannung des ADU festelegen:
analogReference (DEFAULT);  //Referenzspannung des ADU ist die Betriebsspannung

//definierte Ausgangsschaltzustände herstellen:
digitalWrite (PTTU, HIGH);
digitalWrite (PDU, LOW);
digitalWrite (PDV, LOW);
digitalWrite (NFOUT, LOW);
digitalWrite (LED_rtgn, LOW); //LED auf rot stellen
digitalWrite (BYP, LOW); //Bypassrelais stromlos schalten (Bypass aktiv)
digitalWrite (UGATE, LOW); //Gatespannung abschalten
digitalWrite (SRR, LOW); // Sendekontroll-LED abschalten

//Öffnen der seriellen Schnittstellen und kurze Pause:

Serial.begin(9600);	//Hardwareschnittstelle für Debug-Funktion

Ser818V.begin(9600); //Softwareschnittstelle zum DRA-818V
Ser818U.begin(9600); //Softwareschnittstelle zum DRA-818U

delay (100);

squ_v = analogRead(USQ)/128; //Pegeleinstellung für die Rauschsperre messen und in Parameter umrechnen
/*Anmerkung: Nach dem Ändern dieses Spannungspegels muss das Programm neu
  gestartet werden, damit die Änderung wirksam wird. Dazu ist z.B. der 
  kleine Reset-Taster auf dem Arduino-Board zu betätigen.
*/

//Betriebsparameter in den DRA818-V laden:
digitalWrite (PDV, HIGH); //Power-Down-Zustand des DRA818-V aufheben
delay (500); //kurze Einschwingpause

//allgemeine Modulparameter laden
Ser818V.print ("AT+DMOSETGROUP=");
Ser818V.print (GBW_V,1);
Ser818V.print (",");
Ser818V.print (FRX_V,4);  //Sendefrequenz = Empfangsfrequenz, wird hier nur der Vollständigkeit halber übertragen
Ser818V.print (",");
Ser818V.print (FRX_V,4);
Ser818V.print (",");
Ser818V.print (TX_CTCSS_V);
Ser818V.print (",");
Ser818V.print (squ_v);
Ser818V.print (",");
Ser818V.println (RX_CTCSS_V);
delay (100);

//Lautstärke einstellen
Ser818V.print ("AT+DMOSETVOLUME=");
Ser818V.println (VOL_V,1);

//Betriebsparameter in den DRA818-U laden:
digitalWrite (PDU, HIGH); //Power-Down-Zustand des DRA818-U aufheben
delay (500); //kurze Einschwingpause

//allgemeine Modulparameter laden
Ser818U.print ("AT+DMOSETGROUP=");
Ser818U.print (GBW_U,1);
Ser818U.print (",");
Ser818U.print (FTX_U,4);
Ser818U.print (",");
Ser818U.print (FTX_U,4); //Sendefrequenz = Empfangsfrequenz, wird hier nur der Vollständigkeit halber übertragen
Ser818U.print (",");
Ser818U.print (TX_CTCSS_U);
Ser818U.print (",");
Ser818U.print (SQU_U);
Ser818U.print (",");
Ser818U.println (RX_CTCSS_U);
delay (100);

//Lautstärke einstellen
Ser818U.print ("AT+DMOSETVOLUME=");
Ser818U.println (VOL_U,1);


meldungsausgabe ("v10"); //Versionsnummer der Software im Morsecode (nur als Tonausgabe am NF-Ausgang)

}

//*******************************************************************
//*******************Hauptprogrammschleife***************************
//*******************************************************************

void loop() 
  {
  ub_messung ();  //Betriebsspannungskontrolle
  tmp_messung (); //Temperaturkontrolle des Kühlkörpers

  /*Abfrage des Squelch-Ausgangs des VHF-Moduls und Steuerung des PTT-
  Eingangs des UHF-Moduls mit Abschaltverzögerung des Senders
  */
    if ((digitalRead (SQV)) == LOW)   //Squelch des VHF-Moduls ist offen, das Modul emnpfängt auf der eingestellten Frequenz ein Trägersignal
    {
      pa_einschalten (); //Einschalten der PA wenn keine Unterspannungs- oder Übertemperaturmeldung im Statusspeicher vorliegt und der Ausgangspegel hoch genug ist
      digitalWrite (PTTU, LOW); //UHF-Modul auf Senden schalten
      digitalWrite (SRR, HIGH); //Sende-LED einschalten 
      pa_sperre (); //Überwachung der Ausgangsspannung der Endstufe. Die Überschreitung führt zum Setzen von Bit3 der Variable "statusspeicher". Anderenfalls wird das Bit3 gelöscht.
      bitSet (statusspeicher,7);       //PTT-Speicherbit setzen
    }
    else                  //Squelch des VHF-Moduls ist geschlossen, das Modul empfängt keinen Träger auf der eingestellten Frequenz
    {
      if (bitRead(statusspeicher,7) == 1)      //"1" bedeutet, der UHF-Sender war bisher eingeschaltet
      {
        altezeit1 = millis();                 //Zeitgeber für die Umschaltverzögerung TX->RX starten
        bitClear (statusspeicher,7);         //PTT-Speicherbit zurücksetzen
      }
      if ((millis() - altezeit1) > TX_DELAY)  //Nach Ablauf der  Verzögerungszeit TX_DELAY wird auf Empfang geschaltet
      {
      digitalWrite (PTTU, HIGH); //UHF-Modul auf Empfang schalten
      //digitalWrite (PDU, LOW); //DRA818-U in den Power-Down-Zustand versetzen
      digitalWrite (SRR,  LOW); //Sende-LED abschalten    
      pa_abschalten ();
      }

  //Rücksetzen der PA-Sperre während des Empfangs und nach der Zeit, die in der Variablen PA_OFF_DELAY hinterlegt ist

      if (bitRead(statusspeicher,3) ==1)
      {
        if ((millis()-altezeit3) > PA_OFF_DELAY) //Nach Ablauf der Verzögerungszeit PA_OFF_DELAY wird die Pegelunterschreitungssperre der PA zurückgesetzt
        {
        bitClear (statusspeicher,3);
        }
      }
    }

    #ifdef debug
      Serial.print("Statusspeicher = ");
      Serial.println(statusspeicher, BIN);
      delay (2000); //hält die Anzeige kurz an, verändert damit aber auch die Programmablaufzeit
    #endif

  //Fehlerauswertung
    //rote und grüne LED schalten
    if ((statusspeicher & 0b01111111) > 0) //Ausblenden (Maskierung) des PTT-Statusbits, um nur Fehlermeldungen zu erfassen
    {
      digitalWrite (LED_rtgn, HIGH); //rote LED "Störung" einschalten, grüne LED "Betrieb" aus
    }
    else
    {
      digitalWrite (LED_rtgn, LOW); //grüne LED "Betrieb" einschalten, rote LED "Störung" aus
    }
    //Zeitgesteuerte Ausgabe von Texten im Morsecode in Abhängigkeit von der Fehlerart
    if (((millis()- altezeit2) > MORSEINTERVALL) && ((statusspeicher  & 0b01111111) > 0))
    {
      altezeit2 = millis(); //Start des Zeitgebers für den Ausgabezyklus der Meldungen
      pa_einschalten (); //Endstufe einschalten 
      digitalWrite (PTTU, LOW); //UHF-Modul auf Senden schalten
      digitalWrite (SRR, HIGH); //Sende-LED einschalten
      pa_sperre (); //HF-Spannungsüberwachung der Endstufe
      delay (1000); // kurz auf die Sendebereitschaft warten

      if (bitRead(statusspeicher,0)>0)
      {
        meldungsausgabe ("bat");  
      }
      if ((bitRead(statusspeicher,1) || bitRead(statusspeicher,2))>0)
      {
        meldungsausgabe ("tmp"); 
      }
      if (bitRead(statusspeicher,3)>0)
      {
        meldungsausgabe ("pa"); 
      }
      digitalWrite (PTTU, HIGH); //UHF-Modul auf Empfang schalten
      pa_abschalten ();   
      digitalWrite (SRR,  LOW); //Sende-LED abschalten    
    }
  }
//*******************************************************************
//*******************Programmfunktionen******************************
//*******************************************************************

//Messung der Betriebsspannung und Umrechnung in Volt (Auflösung: 1V), Setzen des Statusbits bei Betriebsspannungsunterschreitung
void ub_messung (void)
  {
    int u_mess12 = analogRead(U12)/36;
    #ifdef debug
    Serial.print("Versorgungsspannung = ");
    Serial.println(u_mess12);
    #endif
    if ((bitRead(statusspeicher,0) == 0) && (u_mess12 < UB12_MIN))  //Fehlerbit setzen beim Unterschreiten der minimalen Betriebsspannung
      {
        bitSet (statusspeicher,0);
      }
   if ((bitRead(statusspeicher,0) > 0) && (u_mess12 >= (UB12_MIN + UB12HYST)))  //Fehlerbit rücksetzen beim Überschreiten der minimalen Betriebsspannung plus Hysteresewert
      {
        bitClear (statusspeicher,0);
      }
  }

//Umschalten auf Empfang und Abschaltung der Endstufe
void pa_abschalten ()
  {
    digitalWrite (UGATE, LOW); //Gatespannung der Endstufe abschalten
    delay (100); //kurze Wartezeit
    digitalWrite (BYP, LOW); //Bypass umschalten, Endstufe aus dem Signalweg nehmen
  }

//Umschalten auf Senden und Zuschaltung der Endstufe
void pa_einschalten ()
  {
  if ((bitRead(statusspeicher,0) + bitRead(statusspeicher,2) + bitRead(statusspeicher,3)) == 0) //PA wird nur zugeschaltet, wenn keine Unterspannung oder Übertemperatur vorliegt.
    {
    digitalWrite (BYP, HIGH);//Bypass umschalten, Endstufe in den Signalweg legen
    delay (100); //kurze Wartezeit
    digitalWrite (UGATE, HIGH); //Gatespannung der Endstufe zuschalten
    }
  }

//Messung der Ausgangsspannung am PA-Ausgang, Sperrung der PA bei Unterschreitung einer Mindestspannung (-leistung)
void pa_sperre ()
  {
    delay (500); //Messzeit für HF-Spannung
    int usend_mess = analogRead(USEND);
    #ifdef debug
      Serial.print("ADU-Wert HF-Spannung= ");
      Serial.println(usend_mess);
    #endif
    if (usend_mess < PA_UMIN) //Spannungsschwelle für etwa 2W
      {
        pa_abschalten (); //PA-Modul wird abgeschaltet und es wird via Bypass nur mit Treiberleistung gesendet
        bitSet (statusspeicher,3);
        altezeit3 = millis(); //Start des Zeitgebers für die automatische Löschung des Fehlers
      }
  }

/*Messung der PA-Kühlkörpertemperatur, Umrechnung in Grad Celsius und Auswertung.
 Beim Überschreiten des ersten Schwellwerts wird das Bit1 der Variable "statusspeicher"
 gesetzt, beim Überschreiten des zweiten Werts wird das Bit2 gesetzt.
 Anderenfalls werden die entsprechenden Bits gelöscht.
 */
void tmp_messung (void)
  {
    int kktmp_c = (TMP_BEZ -  analogRead(UTMP))/10; //ADC-Wert auslesen und Umrechnung in Grad Celsius
    #ifdef debug
      Serial.print("Kühlkörpertemperatur = ");
      Serial.println(kktmp_c);
    #endif
    if (kktmp_c > TMP_S1)  //erste Temperaturschwelle, erhöhte KK-Temperatur
    {
      bitSet (statusspeicher,1);
    }
    else
    {
      bitClear (statusspeicher,1);
    }
    if (kktmp_c > TMP_S2)  //zweite Temperaturschwelle, Übertemperatur-Bit setzen 
    {
      bitSet (statusspeicher,2);
    }
    if (kktmp_c < (TMP_S2-TMPHYST)) // Rücksetzen des Übertemperatur-Bits beim Unterschreiten der zweiten Temperaturschwelle abzüglich Hysterese
    {
      bitClear (statusspeicher,2);
    }
  }

//Meldungsausgabe in Morsezeichen
void meldungsausgabe (String meldung)
  {
  int x = meldung.length ();
  for (int i = 0; i < x; i++)
    {
    char y = meldung.charAt (i);
    mz_ausgabe (y);
    zeichenpause ();
    }
  wortpause(); 
  }

//Tonausgabe eines Morsepunkts mit drei Zeiteinheiten
void punktausgabe(void)
    {
     tone (NFOUT, TON_HZ);
     delay (DIT_SPEED);
     noTone(NFOUT);
    }
     
//Tonausgabe eines Morsestrichs mit drei Zeiteinheiten
void strichausgabe(void)
    {
     tone (NFOUT, TON_HZ);
     delay (DIT_SPEED*3);
     noTone(NFOUT);
    }

//Punkt-Strich-Abstand
void p_s_abstand(void)
    {
     delay(DIT_SPEED);
    }
//Zeichenpause
void zeichenpause(void)
    {
     delay(DIT_SPEED*3);
    }
//Wortpause
void wortpause(void)
    {
     delay(DIT_SPEED*5);
    }
     
// Ausgabe eines Morsezeichens
void mz_ausgabe(char mzeichen)
{
int code_nr; //Zeiger auf Zeichen in Morsecodetabelle
int x, y; //Hilfsvariablen
//Zeichencodetabelle für 0...9, a...z
int const mz_code[36] = {B00111111, B00111110, B00111100, B00111000, B00110000, B00100000, B00100001, B00100011, 
B00100111, B00101111, B00000110, B00010001, B00010101, B00001001, B00000010, B00010100, B00001011, B00010000, B00000100, 
B00011110, B00001101, B00010010, B00000111, B00000101, B00001111, B00010110, B00011011, B00001010, B00001000, B00000011, 
B00001100, B00011000, B00001110, B00011001, B00011101, B00010011};
//übergebenes Zeichen  in einzelnes ASCII-Byte umwandeln:
y = mzeichen;
 //übergebenes Zeichen prüfen:
 if (y >= 48 && y <= 57)  //Ziffern 0...9 abfangen
    {
      code_nr = y - 48;
    }
    else
    if (y >= 97 && y <= 122)   //Buchstaben a...z abfangen
    {   
      code_nr = y - 87;
    }
    else
     return;            //Rücksprung bei ungültigem Zeichen

x = mz_code[code_nr];

while (x > 1)
  {
      if (bitRead (x,0) == 0)
      {
      punktausgabe();
      }
      else
      {
      strichausgabe();
      }
   x = x >> 1;
   p_s_abstand();  
  }
}
